--[[
   Copyright (c) 2012-2021 baby-bus.com

   TODO:    地图管理类
   Author:  Neighbour Mr.Wang
   Date:    2021.08.04
   
   http://www.baby-bus.com/
   
]]
----------------------
-- 类
----------------------
local M     = classNode("MapManager")
M.TAG       = "MapManager"

----------------------
-- 公共参数
----------------------
-- [常量]

-- 需要控制的一圈地图相对标识
local ControlMaps = {
    cc.p(2, 1),
    cc.p(2, 0),
    cc.p(2, -1),
    cc.p(0, 1),
    cc.p(1, 1),
    cc.p(1, 0),
    cc.p(1, -1),
    cc.p(0, -1),
    cc.p(-1, -1),
    cc.p(-1, 0),
    cc.p(-1, 1), 
}

local Config = require("app.biathlon.Config")

----------------------
-- 构造方法
----------------------
--[[--

构造方法，定义结点实例初始化逻辑

### Parameters:
-   table **params**    参数集合

### Return: 
-   object              对象实例

]]
function M:ctor(params)
    -- [超类调用]
    M.super.ctor(self, params)

    -- 传参数[层]
    self._layer      = params.layer
    -- 单元地图大小
    self._mapSize         = cc.size(1024, 626)
    -- 地图行数
    self._row             = 40
    -- 地图列数
    self._col             = 3
    --地图大小
    self._mapSizeTotal    = cc.size(self._mapSize.width * self._row, self._mapSize.height * self._col)
    -- 地图单元表
    self._cellItem        = {}
    -- 显示的地图单元
    self._cellShow        = {}
    -- 当前地图焦点位置
    self._focusMap        = cc.p(1, 2)
    -- 当前角色所在地图id
    self._currentMapId    = self._focusMap.x + (self._focusMap.y - 1) * self._row 
    -- 是否有边界
    self._hasBound        = true
    self._roleList        = {}
    -- 道具障碍物表
    self._propBarrierList = {}
    -- 缆车列表
    self._cableCarList    = {}
    -- 打靶点列表
    self._shootSiteList   = {}
    -- 路障表
    self._roadBlockList   = {}
    -- 靶场节点列表
    self._shootNodeList   = {} 
    -- 是否开启冲线检测
    self._openLineCheck   = false 
    -- 是否已经有角色赢得了胜利
    self._hasRoleWin      = false
    -- 屏幕人物剧中定位时的偏移量
    self._screenOffX       = 0
end

----------------------
-- 结点渲染 
----------------------
--[[--

实体渲染，处理实体结点加载、事件绑定等相关操作

]]
function M:onRender()
    -- [超类调用]
    M.super.onRender(self)
    self:setContentSize(self._mapSizeTotal)
    -- 加载地图
    self:loadMaps()
    -- 加载物理世界
    self:loadPhysicsWorld()
    -- 加载角色
    self:loadRole()
    -- 加载ai角色
    self:loadAIRole()
    -- 加载道路背景
    self:loadWayBg()
    -- 加载背景装饰
    self:loadBgDecorate()
    -- 加载起点
    self:loadStartPoint()
    -- 加载终点
    self:loadEndPoint()
    -- 加载打靶地点
    self:loadShootSite()
    -- 加载路障
    self:loadRoadblock()
    -- 加载赛道管理器(障碍物, 道具的加载)
    self:loadTrackManager()
    -- 创建障碍物
    self._trackManager:createBarrier()
    -- 创建道具盒子
    self._trackManager:createPropBox()
    -- 加载加速节点
    self:loadSpeedNode()
    -- 加载加速节点
    self:loadSpeedNode(cc.p(9200*2, 24))
end

-- 加载路障
function M:loadRoadblock()
    for i,pos in ipairs(Config.roadBlockPoints) do
        local z = i %2 == 1 and 3980 or 2300
        -- 路障
        local roadblock = import("app.biathlon.node.Roadblock").new({}):p(pos):to(self, z):anchor(cc.p(0.5, 0)):scale(0.8 - 0.02*i)
        table.insert(self._roadBlockList, roadblock)
    end
end

-- 加载起点
function M:loadStartPoint()
    -- 对应的地图id
    local mapId  = self:getMapIdAccordX(V.cw)-1
    -- 加载冰人
    self:loadIceMan(1, 1, cc.p(440, 360), self._cellItem[mapId])
    self:loadIceMan(2, 1, cc.p(440, -150), self._cellItem[mapId])
end

-- 加载终点
function M:loadEndPoint()
    -- 终点线
    local endLine = D.img("biathlon/endpoint/line/1.png"):p(37442, self._mapSize.height+196 + 17*2):to(self, 100):anchor(cc.p(0, 0.5)):scaleY(0.82)
    endLine._x    = 37442
    self._endLine = endLine

    -- 对应的地图id
    local mapId  = self:getMapIdAccordX(37442)-1
    -- 加载冰人
    self:loadIceMan(1, 2, cc.p(604, 390 + 17*2), self._cellItem[mapId])
    self:loadIceMan(2, 2, cc.p(604, -120 + 17*2), self._cellItem[mapId])
end

-- 加载打靶地点
function M:loadShootSite()
    local posList = {cc.p(14234, 603), cc.p(24996, 483),}

    for i,pos in ipairs(posList) do
        -- 打靶点
        local shootSite = D.img("biathlon/decorate/shootsite/3.png"):p(pos):to(self, 17000 - 8000*i):flipY(i == 2)
        shootSite._x    = posList[i].x
        table.insert(self._shootSiteList, shootSite)
    end
end

-- 加载冰人
function M:loadIceMan(iceManType, posType, pos, parent)
    -- 冰人
    local iceMan = import("app.biathlon.node.IceMan").new({
        iceManType = iceManType,
        posType    = posType,
    }):to(parent, 1000):p(pos)
end

-- -- 加载加速节点
function M:loadSpeedNode(offset)
    offset = ifnil(offset, cc.p(0, 0))
    -- 加速节点位置表
    local posList       = Config.speedNodePosList 
    -- 特殊加速节点信息表
    local infoList      = Config.speedNodeInfoList 
    local shootsiteList = {} -- 靶场列表

    -- 需要旋转的的加速节点
    for k,info in pairs(infoList) do
        -- 加速节点
        local speedNode = import("app.biathlon.node.SpeedNode").new({
            imgName     = "biathlon/speedline/1.png",
        }):p(cc.pAdd(info[1], offset)):to(self, 50):scale(0.9):rotate(info[2])

        local offYId = k%4
        offYId = offYId == 0 and 4 or offYId
        speedNode._offY = 85*offYId - 25
        -- 将节点加入道具障碍物表
        self:addNodeToPropBarrierList(speedNode)
    end
    -- 有规律的加速节点
    for k,pos in pairs(posList) do
        for i = 1, 4 do
            -- 加速节点
            local speedNode = import("app.biathlon.node.SpeedNode").new({
                imgName     = "biathlon/speedline/1.png",
            }):p(pos.x + offset.x, pos.y + 85*(i-1) + offset.y):to(self, 50):scale(0.9)

            speedNode._offY = 85*i - 25
            -- 将节点加入道具障碍物表
            self:addNodeToPropBarrierList(speedNode)
        end
    end
end

-- 加载地图
function M:loadMaps()
    local mapId = 0
    for i = 1, self._col do
        local z = i == 2 and 10 or 1
        for j = 1, self._row do
            mapId = mapId + 1
            local map = require("app.biathlon.node.Map").new({
                mapId       = mapId, 
                mapSize     = self._mapSize,
                rowTag      = j,
                colTag      = i,
            }):p(self._mapSize.width * (j - 1), self._mapSize.height * (i - 1)):to(self)
            map:z(z)
            table.insert(self._cellItem, map)
            table.insert(self._cellShow, mapId)
            -- 粒子
            local particle1 = P.newParticle("particle/biathlon/snow.plist", map:px(), map:py() + V.h):to(self, 10000)
            particle1:setPositionType(kCCPositionTypeGrouped)
            local particle2 = P.newParticle("particle/biathlon/snow2.plist", map:px(), map:py() + V.h):to(self, 10000)
            particle2:setPositionType(kCCPositionTypeGrouped)

            map._particle1 = particle1
            map._particle2 = particle2
        end
    end
end

-- 加载道路背景
function M:loadWayBg()
    -- 道路背景位置列表
	local posList   = Config.wayBgPosList
    -- 有需要补图的地图块(编号: 位置)
    local extraInfo = Config.extraWayInfoPosList

    local wayBgList = {}
    for i,pos in ipairs(posList) do
        -- 对应的地图id
        local mapId  = self:getMapIdAccordX(pos.x)
        local mapPos = self:convertToWorldSpace(pos)
        mapPos       = self._cellItem[mapId]:convertToNodeSpace(mapPos)
        -- 道路背景
        local wayBg = D.img("biathlon/ways/"..i..".png"):p(mapPos):to(self._cellItem[mapId], 100):anchor(cc.p(0, 0)):scale(1.2505)
        if extraInfo[tostring(i)] then
            local pos = extraInfo[tostring(i)]
            local wayBgExtra = D.img("biathlon/ways/"..i.."_1.png"):p(pos.x/1.25, pos.y/1.25):to(wayBg):anchor(cc.p(0, 0))
        end
        if i >= 2 then -- 补充一些道路(原有道路复制到右边)
            for j = 1, 3 do
                -- 对应的地图id
                local mapId  = self:getMapIdAccordX(pos.x+9216 * j)
                local mapPos = self:convertToWorldSpace(cc.p(pos.x + 9201*j, pos.y + 17*j))
                mapPos       = self._cellItem[mapId]:convertToNodeSpace(mapPos)
                -- 道路背景
                local wayBg2 = wayBg:clone():p(mapPos):to(self._cellItem[mapId], 100)
                if i == 10 and j == 3 then -- 含终点的道路
                    wayBg2:display("biathlon/ways/11.png")
                    -- 道路
                    local way = D.img("biathlon/ways/10.png"):p(wayBg2:point()):to(self._cellItem[mapId+1], 50):anchor(cc.p(0, 0)):scale(1.25)
                end
            end
        end
    end
end

-- 根据X坐标获得对应的地图id
function M:getMapIdAccordX(x)
	return math.ceil(x/self._mapSize.width) + self._row+1
end

-- 加载背景装饰
function M:loadBgDecorate()
    -- 加载栅栏1
    self:loadFences(Config.ordinaryFencePosList, 1)
    -- 加载栅栏3
    self:loadFences(Config.specialFencePosList, 3)
    -- 加载树
    self:loadTrees()
    -- 加载缆车
    self:loadCablecar(cc.p(-456, 6))
    self:loadCablecar(cc.p(8804, 6), true)
    self:loadCablecar(cc.p(8804+9200*2, 6), true)
    -- 加载冰山
    self:loadMountains()
    -- 加载靶场
    self:loadRange()
    -- 加载帐篷
    self:loadTent()
end

-- 加载帐篷
function M:loadTent()
    -- 帐篷位置表
    local posList = {}
    T.copyTable(posList, Config.tentPosList)
       
    for k,pos in pairs(posList) do
        -- 对应的地图id
        local mapId  = self:getMapIdAccordX(pos.x+400)-1 -- (+400保证靠右的层级在上面)
        local mapPos = self:convertToWorldSpace(pos)
        mapPos       = self._cellItem[mapId]:convertToNodeSpace(mapPos)
        -- 帐篷
        local tent = D.img("biathlon/decorate/tent.png"):p(mapPos):to(self._cellItem[mapId], 6000-mapPos.y)
    end
end

-- 加载靶场
function M:loadRange()
    -- 靶场信息(类型, 位置)
	local rangeInfo = {
        {2, cc.p(14585, 311)}, {1, cc.p(6941+9200*2, 741)},
    }

    for i,info in ipairs(rangeInfo) do
        -- 对应的地图id
        local mapId  = self:getMapIdAccordX(info[2].x-300) -- (800保证靠右的层级在上面)
        local mapPos = self:convertToWorldSpace(info[2])
        mapPos       = self._cellItem[mapId]:convertToNodeSpace(mapPos)
        -- 靶场
        local shootsite     = D.img("biathlon/decorate/shootsite/"..info[1]..".png"):p(mapPos):to(self._cellItem[mapId], 50):scale(1.05)
        shootsite._mapId    = mapId
        shootsite._worldPos = info[2]

        if info[1] == 1 then
            local fence1 = D.img("biathlon/decorate/fence/2.png"):p(-30, 360):to(shootsite, 5000)
            local fence2 = D.img("biathlon/decorate/fence/2.png"):p(1040, 360):to(shootsite, 5000)
        else
            local fence1 = D.img("biathlon/decorate/fence/2.png"):p(-40, 180):to(shootsite, 5000)
            local fence2 = D.img("biathlon/decorate/fence/2.png"):p(1060, 180):to(shootsite, 5000)
        end
        table.insert(self._shootNodeList, shootsite)
    end
end

-- 加载缆车
function M:loadCablecar(pos, isToSelf)
    local parent = self
    if not isToSelf then
        -- 对应的地图id
        local mapId  = self:getMapIdAccordX(pos.x)+1
        local mapPos = self:convertToWorldSpace(pos)
        -- 地图
        parent = self._cellItem[mapId]
        pos    = parent:convertToNodeSpace(mapPos)
    end

    -- 缆车
    local cableCar = import("app.biathlon.node.CableCar").new({
        imgName = "biathlon/decorate/cablecar/base1.png"
    }):p(pos):to(parent, 7000):anchor(cc.p(0, 0))

    table.insert(self._cableCarList, cableCar)
end

-- 加载冰山
function M:loadMountains()
    -- 冰山位置表
    local mountainPosList = Config.iceMountainInfoList["posList"]
    -- 冰山类型表
    local mountainTypeList = Config.iceMountainInfoList["typeList"]
    -- 冰山大小
    local mountainSizeList = Config.iceMountainInfoList["sizeList"]

    for i,pos in ipairs(mountainPosList) do
        -- 对应的地图id
        local mapId  = self:getMapIdAccordX(pos.x-300)-1
        local mapPos = self:convertToWorldSpace(pos)
        mapPos       = self._cellItem[mapId]:convertToNodeSpace(mapPos)
        -- 冰山
        local mountain = D.img("biathlon/decorate/mountain/"..mountainTypeList[i]..".png"):p(mapPos)
        :to(self._cellItem[mapId], 6000-mapPos.y):anchor(cc.p(0.5, 0)):scale(mountainSizeList[i])
    end
end

-- 加载树
function M:loadTrees()
	-- 树位置表
    local treePosList = Config.treesPosList

    for i,pos in ipairs(treePosList) do
        -- 对应的地图id
        local mapId  = self:getMapIdAccordX(pos.x+200)-1 -- (+200保证靠右的层级在上面)
        local mapPos = self:convertToWorldSpace(pos)
        mapPos       = self._cellItem[mapId]:convertToNodeSpace(mapPos)
        -- 树
        local tree = D.img("biathlon/decorate/tree/"..N.random(1, 8)..".png"):p(mapPos):anchor(cc.p(0.5, 0)):scale(N.random(7, 9)*0.1)
        :to(self._cellItem[mapId], 6000-mapPos.y)
    end
end

-- 加载栅栏
function M:loadFences(fencePosList, types)
    for i,pos in ipairs(fencePosList) do
        -- 对应的地图id
        local mapId  = self:getMapIdAccordX(pos.x+200)-1 -- (+200保证靠右的栅栏层级在上面)
        local mapPos = self:convertToWorldSpace(pos)
        mapPos       = self._cellItem[mapId]:convertToNodeSpace(mapPos)
        -- 栅栏
        local fence = D.img("biathlon/decorate/fence/1.png"):p(mapPos):to(self._cellItem[mapId], 6000-mapPos.y)
        local dis, k = self:getHighAndSlope(fence, pos.x, pos.y)
        fence:rotate(-math.deg(math.atan(k)))
    end
end

-- 加载物理世界
function M:loadPhysicsWorld()
    local physicsWorld = require("app.biathlon.node.PhysicsWorld").new({
        layer      = self._layer,
        mapManager = self
    }):to(self, 10)
    self._physicsWorld = physicsWorld
end

-- 加载角色
function M:loadRole()
    local bone = "hx1"
    local player = require("app.biathlon.node.RolePlayer").new({
        file         = "bone/biathlon/"..bone,
        armatureName = bone,
        roleType     = 1,
        name         = bone.."_",
        layer        = self._layer,
        physicsWorld = self._physicsWorld,
        mapManager   = self,
    }):to(self):anchor(cc.p(0.5, 0.35))

    table.insert(self._roleList, player)
    self._player = player
end

-- 加载ai角色
function M:loadAIRole()
    local boneInfo = { -- 骨骼信息(骨骼名, 角色类型)
        {"hx2", 2}, {"hx2", 1}, {"hx1", 2}
    }
    for i = 1,3 do
        local bone = boneInfo[i][1]
        local role = require("app.biathlon.node.RoleAI").new({
            file          = "bone/biathlon/"..bone,
            armatureName = bone,
            roleType     = boneInfo[i][2],
            name         = bone.."_",
            layer        = self._layer,
            physicsWorld = self._physicsWorld,
            mapManager   = self,
            id           = i
        }):to(self):anchor(cc.p(0.5, 0.35))

        table.insert(self._roleList, role)
    end
end

-- 加载赛道管理器
function M:loadTrackManager()
	-- 赛道管理器
    local trackManager = import("app.biathlon.node.TrackManager").new({
        roleList   = self._roleList,
        mapManager = self
    }):to(self)

    self._trackManager = trackManager
end

--------------------------------------------
-- 功能方法
--------------------------------------------
-- 开始游戏事件
function M:gameStartEvent()
    if self._cableCarList and not tolua.isnull(self._cableCarList[3]) then
        self._cableCarList[3]:hide() -- 隐藏第三个缆车(因为太长了, 不跟随子地图隐藏)
        A.pauseTarget(self._cableCarList[3]) -- 暂停动作
    end

    for i,v in ipairs(self._propBarrierList) do
        v:hide()
        if v._relatedPropList and T.isNotEmpty(v._relatedPropList) then
            for _,child in ipairs(v._relatedPropList) do
                child:hide()
            end
        end
    end
end

-- 将节点加入道具障碍物表
function M:addNodeToPropBarrierList(node)
    if node:px() >= 2*V.w then
        for i,prop in pairs(self._propBarrierList) do
            if math.abs(prop:px() - node:px()) <= 200 then
                if not prop._relatedPropList then -- 增加关联道具表(减少碰撞检测遍历次数)
                    prop._relatedPropList = {}
                end
                table.insert(prop._relatedPropList, node)
                return
            end
        end
    end

    table.insert(self._propBarrierList, node)
end

-- 移除缆车
function M:removeCableCar()
    -- 移除缆车
    local firstCableCar = self._cableCarList[1]
    T.removeOrder(self._cableCarList, firstCableCar)
    firstCableCar = CommonFun.remove(firstCableCar)
end

-- 屏幕晃动
function M:screenShake(dir, offsetX)
    if self._screenShakeAct then return end

    self._screenShakeAct = A.cycle({
        {"delay", 1/60},
        {"fn", function()
            self._screenOffX = self._screenOffX + dir

            if math.abs(self._screenOffX - offsetX) < math.abs(dir) then
                self:stopAction(self._screenShakeAct)
                self._screenShakeAct = nil
                if math.abs(offsetX) <= 1 then
                    self._screenOffX = 0
                else
                    -- 屏幕晃动
                    self:screenShake(-dir, -offsetX/3*2)
                end
            end
        end},
    }):at(self)
end

--------------------------------------------
-- 打靶
-------------------------------------------
-- 打靶开始
function M:shootStart()
    local layer = self._layer -- 层
    if layer._isFinish or (layer._countDown and layer._countDown._isFinish) then return end

    layer._isShoot  = true
    layer._canTouch = false    -- 关闭触控
    -- 关闭移动监听
    layer:stopSchedule()
    -- 移除缆车
    self:removeCableCar()
    self:plan(function()
        if layer._isFinish or (layer._countDown and layer._countDown._isFinish) then return end
        -- 播放音效[sfx26233003]由赛道镜头移动到靶场的音效
        soundEffect:playEffectsfx26233003()
        if self._player._gameStage == 2 then
            -- 移除缆车(移除两个缆车)
            self:removeCableCar()
            -- 第一次打靶开始
            self:firstShootStart()
        else
            self._cableCarList = nil
            -- 第二次打靶开始
            self:secondShootStart()
        end
    end, 0.5)
end

-- 第一次打靶开始
function M:firstShootStart()
    local pos = self._shootNodeList[1]._worldPos
    CommonFun:changeAnchorP(self._shootSiteList[1], cc.p(0.5, 0))
    -- 打靶提示弹动
    self:shootHintSpring(self._shootSiteList[1])

    A.line({-- 镜头移到靶场
        {"delay", 0.4},
        {"easing","sineout",{"moveTo", 1.2, cc.p(-pos.x + V.cw, -pos.y + V.ch)}},
        {"fn", function()
            -- 过渡并加载打靶弹窗
            self:transitionAndLoadShootPop(1)
        end},
    }):at(self)
end

-- 第二次打靶开始
function M:secondShootStart()
    local pos   = self._shootNodeList[2]._worldPos

    CommonFun:changeAnchorP(self._shootSiteList[2], cc.p(0.5, 1))
    -- 打靶提示弹动
    self:shootHintSpring(self._shootSiteList[2])

    A.line({ --  镜头移到靶场
        {"delay", 0.4},
        {"easing","sineout",{"moveTo", 1.2, cc.p(-pos.x + V.cw, -pos.y + V.ch)}},
        {"fn", function()
            -- 过渡并加载打靶弹窗
            self:transitionAndLoadShootPop(2)
        end},
    }):at(self)
end

-- 过渡并加载打靶弹窗
function M:transitionAndLoadShootPop(shootTime)
    local layer = self._layer -- 层
	-- 白屏过渡
    layer:whiteScreen(function ()
        if isfunction(layer.hideCountDown) then
            -- 隐藏倒计时
            layer:hideCountDown()
        end
        if self._player._gameStage ~= 2 then -- 第二次打靶
            self._shootNodeList[2]:display("biathlon/decorate/shootsite/1_1.png")  -- 变为无靶子的
        end
        -- 加载射击弹窗
        layer:loadShootPopup(shootTime)
    end)
end

-- 打靶提示弹动
function M:shootHintSpring(shootHint)
    if tolua.isnull(shootHint) then return end
    A.cycle({
        {"scaleTo", 0.2, 1.12, 0.9},
        {"scaleTo", 0.12, 1, 1},
        {"scaleTo", 0.09, 0.9, 1.12},
        {"scaleTo", 0.08, 1, 1},
        {"scaleTo", 0.08, 1.1, 0.9},
        {"scaleTo", 0.07, 1},
    }, 2):at(shootHint)
end

-- 打靶完成
function M:shootFinish()
    local layer = self._layer -- 层
    if isfunction(layer.showCountDown) then
        -- 显示倒计时
        layer:showCountDown()
    end
    layer._isShoot = false
    -- 打靶成功事件
    self._player:shootSuccessEvent()

	if self._player._gameStage == 2 then
        self._shootSiteList[1]:show() -- 显示
        -- 第一次打靶完成
        self:firstShootFinish()
    else
        self._shootNodeList[2]:display("biathlon/decorate/shootsite/1.png")  -- 变为有靶子的
        self._shootSiteList[2]:show() -- 显示
    end
    -- 镜头恢复继续游戏
    self:cameraResetContinueGame()
end

-- 第一次打靶完成
function M:firstShootFinish()
    -- 第二个缆车开始动
	self._cableCarList[1]:show()
    A.resumeTarget(self._cableCarList[1])
end

-- 镜头恢复继续游戏
function M:cameraResetContinueGame()
    local layer = self._layer -- 层
	
    local y      = V.ch - (40 + self._player._relBodyOffsetY*0.1)*Y_FACTOR
    local mapPos = self:point()
    local pos    = playerFocusOnMap(self._player, self, cc.p(V.w*0.382 + self._screenOffX, y))
    self:p(mapPos)

    -- 镜头移到人物
    A.line({
        {"moveTo", 1.2, pos},
        {"fn", function()
            -- 开启移动监听
            layer:startSchedule()
            -- 打靶完成继续游戏
            self:shootFinishContinue()
        end},
    }):at(self)
    layer:whiteScreen(nil, 0)
end

-- 打靶完成继续游戏
function M:shootFinishContinue()
    if  tolua.isnull(self) or tolua.isnull(self._player) then return end
    local layer = self._layer -- 层

    local player = self._player
    -- 重置所有状态
    player:resetAllState()-- 防止可能出现的bug, 恢复所有状态

    player._extraSpeed = player._extraMaxSpeed/2 -- 给一半的额外速度
    local result     = player:applySkiSpeed(40)
    if result then
        player._fsm:doEvent("toSpeedup")-- 滑行
    end
    layer:plan(function()
        if not layer._isFinish then
            layer._canTouch = true    -- 开启触控
        end
    end, 0.2)
end
--------------------------------------------
-- 碰撞
--------------------------------------------
-- 碰撞判断
function M:crashCheck()
    -- 人物碰撞检测
    self:roleCrashCheck()
    -- 道具障碍碰撞检测
    self:propBarrierCrashCheck()
    -- 靶场检测
    self:shootSiteCheck()
    if self._openLineCheck then
        -- 终点检测
        self:endPointCheck()
    end
end

-- 终点检测
function M:endPointCheck()
    -- 获得角色列表
    local roleList  = self._roleList
    local endLine   = self._endLine -- 终点线

    for k,role in pairs(roleList) do
        if not role._isShootOrPass and role:px() >= endLine._x - 400 then -- 终点前操作
            -- 换赛道冲线
            self._trackManager:changeTrackToEndLine(role)

            if role:px() >= endLine._x - 30 then -- 到达终点
                -- 到终点事件
                self:arriveEndLineEvent(role, endLine)
            end
            if role._isPlayer and not role._gameOver then
                role._gameOver  = true
                local layer     = self._layer -- 层
                layer._canTouch = false    -- 关闭触控

                -- 播放音效[sfx26233011]到达终点后的激励音效
                soundEffect:playEffectsfx26233011()
                sound.stopMusic()
                -- 播放音效[sfx26233010]两侧喷出礼花
                soundEffect:playEffectsfx26233010()
            end
        end
    end
end

-- 到终点事件
function M:arriveEndLineEvent(role, endLine)
	role._isShootOrPass = true
    -- -- 冲向终点线
    role:playSprint()
    if role._rank == 3 and not self._layer._isFinish then -- 第三名产生但游戏未结束
        -- 加载倒计时
        role._layer:loadCountDown()
    end

    if role._isPlayer then
        A.line({
            {"easing","sineout",{"scaleto", 2, 1.2}},
        }):at(self)
    end
    self:plan(function()
        role._body:SetLinearDamping(2)
    end, 0.2)
    -- 没有角色赢得了胜利
    if not self._hasRoleWin then
        if role._isPlayer or role:isInScreen() then
            -- 播放音效[sfx26233009]撞线
            soundEffect:playEffectsfx26233009()
        end
        self._hasRoleWin = true
        local particle1 = P.newParticle("particle/biathlon/chongzhuangxian.plist", endLine:px(), endLine:py()):to(self, 10000)
        A.line({
            {"image", "biathlon/endpoint/line/", 10, 0.08},
        }):at(endLine)
    end
end

-- 靶场检测
function M:shootSiteCheck()
    -- 角色列表
    local roleList = self._roleList
    for k,role in pairs(roleList) do
        if role._gameStage <= 2 then -- 处于前往打靶阶段
            if role:px() >= self._shootSiteList[role._gameStage]._x - 400 then -- //换到对应赛道
                if not role._isShootOrPass then
                    role._isShootOrPass = true    -- 正在打靶
                    local v             = role._body:GetLinearVelocity()
                    if v.x < 20 then v.x = 20 end
                    role._body:SetLinearVelocity(b2Vec2(v.x, v.y-5))
                    -- 换赛道打靶
                    self._trackManager:changeTrackToShoot(role)
                end
             
                if role:px() >= self._shootSiteList[role._gameStage]._x then
                    -- 开始打靶事件
                    role:startShootEvent()
                    
                    if role._isPlayer then
                        -- 打靶开始
                        self:shootStart()
                    else
                        -- 打靶计时
                        role:shootTimeCount()
                    end
                end
            end
        end
    end
end

-- 道具障碍碰撞检测
function M:propBarrierCrashCheck()
	for k,v in pairs(self._propBarrierList) do
        if isnumber(v._coldTime) and v._coldTime > 0 then -- 延迟检测
            v._coldTime = v._coldTime - 1
        else
           -- 道具障碍碰撞检测逻辑
           self:propBarrierCrashCheckLogic(v)
        end
    end
end

-- 道具障碍碰撞检测逻辑
function M:propBarrierCrashCheckLogic(node)
    local disX = 100000 -- 角色与道具的X轴距离
    for i,role in pairs(self._roleList) do
        local dis = PT.distance(role:point(), node:point()) -- 角色与道具的距离
        -- 人物与道具/障碍物接触
        if math.abs(role._relBodyOffsetY - node._offY) < 30 then -- 基本在同一水平线
            if dis < 100 and role._floorDis <= node._highLimit then
                -- print("❤️撞到障碍物/道具了")
                node:beCrashEvent(role) -- 碰撞事件
                disX = 60               -- 为了防止做延迟检测
                break
            -- ai接近障碍物, 且玩家领先ai, ai绕过障碍物
            elseif dis < 310 and dis > 120 and node._isBarrier then
                -- 角色接近障碍物逻辑
                self:roleNearBarrierLogic(role, dis)
            end
        end
        
        if disX > node:px() - role:px() and (node:px() - role:px()) > -V.w*2 then -- 判断 角色未超过障碍物两屏
            disX = node:px() - role:px()
        end
    end

    if disX == 100000 then -- 说明滞后了最后一名V.w*2像素距离直接删
        T.removeOrder(self._propBarrierList, node)
        node:hide() -- 移除时太消耗性能, 改为隐藏
    elseif disX > V.w * 2 then -- 离人物太远, 隐藏并延迟检测距离
        node._coldTime = disX/V.w * 30
        node:hide()
    elseif not node:isVisible() then
        node:show()
        if node._relatedPropList and T.isNotEmpty(node._relatedPropList) then -- 将关联的节点加入碰撞判断表
            for _,child in ipairs(node._relatedPropList) do
                child:show()
                table.insert(self._propBarrierList, child)
            end
        end
        node._relatedPropList = nil
    end
end

-- 角色接近障碍物逻辑
function M:roleNearBarrierLogic(role, dis)
	if role._isAI  and dis < 190 and self._player:px() > role:px() then
        -- 更换赛道
        self:getTrackManager():changeTrack(role, N.randomExcept(-1, 1, 0))
    elseif role._isPlayer and role._floorDis < 10 then
        local body = role._body
        -- 获得线速度
        local vel  = body:GetLinearVelocity()
        local vecX = vel.x/15*14
        vecX = vecX > 3 and vecX or 3

        body:SetLinearVelocity(b2Vec2(vecX, vel.y))
    end
end

-- 人物碰撞检测
function M:roleCrashCheck()
	-- 获得角色列表
    local roleList  = self._roleList
    local otherRole = nil

    for k,role in pairs(roleList) do
        for i = 1, 4 do
            otherRole = roleList[i]
            if math.abs(role._relBodyOffsetY - otherRole._relBodyOffsetY) < 30 then
                local disX = role:px() - otherRole:px()

                if  disX > 0 and disX < 80 then -- 距离小于80判断是否要躲避, 或者是否已经碰撞
                    if disX < 70 and PT.distance(role:point(), otherRole:point()) < 70 then
                        -- 人物碰撞事件
                        self:roleCrashEvent(role, otherRole)
                        break
                    elseif disX >= 74 and disX < 76 then
                        if role._isPlayer then
                            -- 更换赛道
                            self:getTrackManager():changeTrack(otherRole, N.randomExcept(-1, 1, 0))
                        elseif otherRole._isPlayer then
                            -- 更换赛道
                            self:getTrackManager():changeTrack(role, N.randomExcept(-1, 1, 0))
                        end
                        break
                    end
                end
            end
        end
        if role._isAI then
            -- 角色位置提示
            role:rolePosTips(role)
        end
    end
end

-- 人物碰撞事件
function M:roleCrashEvent(role, otherRole)
	if not role._ignoreSynPhyPos and not otherRole._ignoreSynPhyPos then-- 未忽略物理世界同步
	    if not role._rocketAct and not otherRole._rocketAct then  -- 没有在坐火箭
            if not role._isShootOrPass and not otherRole._isShootOrPass then -- 不在打靶
                local velA = role._body:GetLinearVelocity()
                local velB = otherRole._body:GetLinearVelocity()

                if not role._isCrash then
                    local vx = velA.x < 14 and velA.x or 14
                    role._isCrashFront = true -- 碰撞时处于前方
                    -- 被撞击事件
                    role:beCrashEvent(vx+9)
                end

                if not otherRole._isCrash then
                    local vx = velB.x*0.6 < 9 and velB.x*0.6 or 9
                    -- 被撞击事件
                    otherRole:beCrashEvent(vx-2)
                end
            end
        end
    end
end

--------------------------------------------
-- 地图相关
--------------------------------------------
-- 获得离地面高度及地面斜率
function M:getHighAndSlope(spr, sprX, sprY)
    -- 设置精灵对应地形边缘点的id
    self:setSprEdgePosId(spr, sprX)

    local endP   = Config.edgePoints[spr._edgePosId + 1]
    local begP   = Config.edgePoints[spr._edgePosId]
    -- 计算点到直线的距离和斜率
    local dis, k = CommonFun:disPoint2Line(sprX, sprY, begP.x, begP.y, endP.x, endP.y)
 
    return dis, k
end

-- 设置精灵对应地形边缘点的id
function M:setSprEdgePosId(spr, sprX)
    if not spr._edgePosId then
        spr._edgePosId = 2
    else
        local edgePos   = Config.edgePoints[spr._edgePosId] -- 记录的边缘点
        if sprX < edgePos.x then
            spr._edgePosId = spr._edgePosId - 12 -- 当前位置小于上次记录的, 说明倒退了, 从滞后的点开始遍历
        end
        spr._edgePosId = spr._edgePosId <= 2 and 2 or spr._edgePosId
    end

    local lineEndId = 2 -- 精灵处于的直线的末端点id
    for i = spr._edgePosId, #Config.edgePoints do
        if Config.edgePoints[i].x > sprX then
            lineEndId = i
            break
        end
    end
    
    spr._edgePosId = lineEndId-1
end

-- 处理地图移动逻辑
function M:judgeMap(role)
    if self._hasBound then
        local y = V.ch - (40 + role._relBodyOffsetY*0.1)*Y_FACTOR
        playerFocusOnMap(role, self, cc.p(V.w*0.382 + self._screenOffX, y))
    else
        -- 循环地图移动
        local playerWorldPos = self:convertToWorldSpace(role:point())
        local focusPos       = cc.p(V.w_2, V.h_2)
        -- 计算出要调整的坐标差值
        local x = self:getPositionX() + focusPos.x - playerWorldPos.x
        local y = self:getPositionY() + focusPos.y - playerWorldPos.y
        self:setPosition(x, y)
    end

    -- 地图二维坐标转换成地图编号
    local index = self:getCurrentMapIdByWorldPos(role:worldpoint())
    if self._currentMapId ~= index then
        -- 调整地图位置
        self:adjustMap(index)
        -- 更新当前中心点地图
        self._currentMapId = index
    end
end

-- 获取当前世界坐标下的地图id
function M:getCurrentMapIdByWorldPos(worldPos)
    local position = self:convertToNodeSpace(worldPos)
    return self:getCurrentMapIdByLocalPos(position)
end

-- 获取地图坐标下的地图id
function M:getCurrentMapIdByLocalPos(localPos)
    local rowTag = math.floor(localPos.x / self._mapSize.width) + 1
    local colTag = math.floor(localPos.y / self._mapSize.height) + 1
    rowTag = self:limitTagArea(rowTag, self._row)
    colTag = self:limitTagArea(colTag, self._col)
    return rowTag + (colTag - 1) * self._row
end

-- 限制标识范围（防止小于1或大于最大标识）
function M:limitTagArea(tag, maxTag)
    tag = (tag - 1) % maxTag + 1
    return tag
end

-- 调整地图位置
function M:adjustMap(mapTag)
    if not mapTag then return end

    local centerMapPos     = self._cellItem[mapTag]:point()
    local currentMapPosTag = self._cellItem[mapTag].mapPosTag
    -- 当前需要显示的地图列表
    local currentMapsTag   = {}

    table.insert(currentMapsTag, mapTag)
    for i,v in ipairs(ControlMaps) do
        local rowTag = self:limitTagArea(currentMapPosTag.x + v.x, self._row)
        local colTag = self:limitTagArea(currentMapPosTag.y + v.y, self._col)
        local tag = rowTag + (colTag - 1) * self._row
        table.insert(currentMapsTag, tag)
        
        self._cellItem[tag]:p(centerMapPos.x + self._mapSize.width * v.x,centerMapPos.y + self._mapSize.height * v.y)
    end

    -- 地图的隐藏和显示
    self:resManager(currentMapsTag)
end

-- 地图的隐藏和显示
function M:resManager(currentMapsTag)
    for i,v in ipairs(currentMapsTag) do
        local hasFind = false
        for j,k in ipairs(self._cellShow) do
            if v == k then
                hasFind = true
                table.remove(self._cellShow, j)
                break
            end
        end
        if not hasFind then
            self._cellItem[v]:showMap()
            -- 显示雪花粒子
            self._cellItem[v]:showSnowPar()
        end
    end
    for i,v in ipairs(self._cellShow) do
        self._cellItem[v]:hideMap()
        -- 隐藏雪花粒子
        self._cellItem[v]:hideSnowPar()
    end

    self._cellShow = currentMapsTag
end

----------------------
-- 公共函数
----------------------
-- 设置目标位置
function M:setTargetPosition()
    -- 地图移动
    self:judgeMap(self._player)
end

-- 获得玩家
function M:getPlayer()
	return self._player
end

-- 获得赛道管理器
function M:getTrackManager()
	return self._trackManager
end

-- 获得角色列表
function M:getRoleList()
	return self._roleList
end

-- 获得物理世界
function M:getPhysicsWorld()
	return self._physicsWorld
end


----------------------
-- 结点析构
----------------------
function M:onDestructor()
    -- [超类调用]
    M.super.onDestructor(self)
end

----------------------
-- 过度动画开始之前的调用
----------------------
--[[--

1.资源释放，退出场景前的操作准备
2.处理视图结点卸载、事件解除绑定等相关操作

]]
function M:onExitTransitionStart()

end

return M