--[[
   Copyright (c) 2012-2021 baby-bus.com

   TODO:    物理世界
   Author:  Neighbour Mr.Wang
   Date:    2021.08.04
   
   http://www.baby-bus.com/
   
]]

----------------------
-- 类
----------------------
local M = classNode("PhysicsWorld")
M.TAG   = "PhysicsWorld"


----------------------
-- 公共参数
----------------------
local Config = require("app.biathlon.Config")

----------------------
-- 构造方法
----------------------
--[[--

构造方法，定义结点实例初始化逻辑

### Parameters:
-   table **params**    参数集合

### Return: 
-   object              对象实例

]]
function M:ctor(params)
    -- [超类调用]
    M.super.ctor(self, params)
    -- 主层
    self._layer      = params.layer
    -- 地图管理器
    self._mapManager = params.mapManager
end

----------------------
-- 结点渲染 
----------------------
--[[--

实体渲染，处理实体结点加载、事件绑定等相关操作

]]
function M:onRender()
    -- [超类调用]
    M.super.onRender(self)
    self:setContentSize(cc.size(0.1, 0.1))
    -- 创建物理世界
    self:createPhysicalWorld()
    --------------------------------------------
    -- test
    ----------------------------------------
    -- ZBox2D.openDebug(self, true)
    -- loadHelpNodeList(self, Config.edgePoints, 1.2)
    -- self:bindTouch()
end

-- 创建人物刚体
function M:createRoleBody(userDate, pos)
    -- 形状信息表
    local shapeInfo = {
        -- 矩形
        shapeType = 5,
        width     = 10,
        height    = 5,
    }
    -- 刚体信息
    local bodyInfo = {
        bodyType       = b2_dynamicBody,
        userData       = userDate,
        point          = pos,
        angularDamping = 0.8,
        linearDamping  = 0,
        density        = 1,
        shape          = shapeInfo,
        fixedRotation  = true,
        friction       = 0.13
    }
    -- 创建刚体, 创建简单的纹理
    local body = ZBox2D.createBody(bodyInfo)
    userDate._body = body
    -- 分组索引
    ZBox2D.setGroupIndex(body, -1)
end

-- 创建物理世界
function M:createPhysicalWorld()
    local emptyNode = U.loadNode({ contentSize = cc.size(0.1, 0.1) }):to(self, 55)
	self._emptyNode = emptyNode
    -- 物理世界边界点
    local edgePoints = Config.edgePoints
    local world   = ZBox2D.createWorldAndEdge(b2Vec2(0.0, -14.2), nil, nil, edgePoints)
    self._world   = world
    -- 刷新
    ZBox2D.update(handlerP(self, self.tick, ZBox2D.getTimeStep()), ZBox2D.getTimeStep(), self._emptyNode)
end

-- 同步精灵的位置跟角度
-- delay:刷新频率
function M:tick(stepTime)
    self._world:Step(stepTime, 4, 4)

    local body = self._world:GetBodyList()
    if not tolua.isnull(body) then
        while body and not tolua.isnull(body) do 
            if body:GetUserData() then
                if body._hasDestroy then
                    self._world:DestroyBody(body)
                else
                   -- 精灵刚体处理
                   self:sprBodyDealWith(body)
                end
            end
            -- 获得下一个body
            body = body:GetNext()
		end
    end

    if self._layer and self._layer._isGameStart then
        -- 碰撞判断
        self._mapManager:crashCheck()
    end
end

-- 精灵刚体处理
function M:sprBodyDealWith(body)
    -- 转换类型为精灵
    local spr  = tolua.cast(body:GetUserData(), "cc.Sprite")
    local x, y = body:GetPosition().x * PTM_RATIO, body:GetPosition().y * PTM_RATIO

    -- 获得角色离地面高度及地面斜率
    local dis, k = self._mapManager:getHighAndSlope(spr, x, y)
    -- 影子角度处理
    self:shadowAngleDeal(spr, body, dis, k)
    -- 道具飞行落地
    spr:propFlyToGround()
    -- 角色角度处理
    self:roleAngleDeal(spr, k)
    local velocity = body:GetLinearVelocity() 
    
    if spr._isPlayer then
        -- 屏幕晃动判断
        self:screenShakeCheck(spr, dis)
        spr:setSpeedParSpeed(velocity.x, velocity.y) -- 设置加速线条速度
    end
    if not spr._ignoreSynPhyPos then -- 不忽略物理世界位置同步
        if not spr._isShootOrPass then
            -- 精灵状态改变
            self:sprStateChange(spr, dis, body, velocity)
        end
        -- 额外速度处理
        self:extraSpeedDeal(spr)
        -- 偏移量添加
        y = self:aadOffsetY(spr, y)
        -- 更新位置、角度，同步物理世界和渲染世界
        spr:setPosition(cc.p(x, y))
        -- 刀痕
        self:loadBlade(spr, x, y, dis, velocity)
        -- 雪痕
        self:loadSnowStrail(spr, dis, velocity)
        self:loadSnowStrail(spr, dis, velocity)
    else
        -- 移除雪痕迹
        self:snowStrailRemove(spr, 0)
    end
    if not spr._ignoreSynPhyZ then-- 不忽略物理世界层级同步
        spr:z(5000 - spr._relBodyOffsetY*10)
    end
end

-- 额外速度处理
function M:extraSpeedDeal(spr)
	if spr._extraSpeed > 5.5 then -- 额外速度减少
        spr._extraSpeed = spr._extraSpeed - spr._extraSpeed/20
	elseif spr._extraSpeed > 0.2 then -- 额外速度减少
        spr._extraSpeed = spr._extraSpeed - spr._extraSpeed/80
    end
    if spr._speedNodeExSpeed > 0 then -- 额外速度减少
        spr._speedNodeExSpeed = spr._speedNodeExSpeed - 0.035
        if spr._isPlayer then
            spr:applySkiSpeed(0)
        end
    end
end

-- 偏移量添加
function M:aadOffsetY(spr, y)
    if spr._relBodyOffsetY then
        y = y + spr._relBodyOffsetY
    end
    if spr._specialOffsetY then -- 特殊情况偏移量(飞行时)
        y = y + spr._specialOffsetY
    end

    return y
end

-- 雪痕
function M:loadSnowStrail(sprite, dis, velocity)
    if tolua.isnull(sprite) or (isfunction(sprite.isInScreen) and not sprite:isInScreen())  then return end
    if not sprite._topSnow then
        if velocity.x > 5 and dis < 7 then
            -- 顶部雪
            local topSnow      = D.img("biathlon/snowstrail/top/1.png"):p(110, 54):to(sprite, 50):anchor(cc.p(1, 0))--:hide()
            -- 底部雪
            local bottomSnow   = D.img("biathlon/snowstrail/bottom/1.png"):p(132, 69):to(sprite, -50):anchor(cc.p(1, 0))
            sprite._topSnow    = topSnow
            sprite._bottomSnow = bottomSnow
            -- 加载出发时的大片雪
            self:loadBigSnowWithSetOut(sprite)
            -- 雪痕动作
            self:snowAct(sprite)
        end
    elseif sprite._topSnow then
        if velocity.x < 4 or dis > 10 then
            -- 移除雪痕迹
            self:snowStrailRemove(sprite, dis)
        elseif velocity.x < 14 then
            local sc = velocity.x/18 + 0.2
            sprite._bottomSnow:scale(sc)
            sprite._topSnow:scale(sc)
        else
            sprite._bottomSnow:scale(1)
            sprite._topSnow:scale(1)
        end
    end
end

-- 雪痕动作
function M:snowAct(sprite)
    A.line({
        {"fn", function()
            A.one({"image", "biathlon/snowstrail/bottom/", 3, 0.04}):at(sprite._bottomSnow) 
        end},
        {"image", "biathlon/snowstrail/top/", 3, 0.04},
        {"fn", function()
            A.cycle({
                {"fn", function()
                    A.one({"imagerange", "biathlon/snowstrail/bottom/", 3, 4, 0.06}):at(sprite._bottomSnow)  
                end},
                {"imagerange", "biathlon/snowstrail/top/", 3, 4, 0.06},
            }):at(sprite._topSnow)
        end},
    }):at(sprite._topSnow)
end

-- 加载出发时的大片雪
function M:loadBigSnowWithSetOut(sprite)
	-- 大片雪
    local bigSnow1 = D.img("biathlon/snowstrail/big.png"):p(100, 69):to(sprite, 150):anchor(cc.p(1, 0))
    local bigSnow2 = D.img("biathlon/snowstrail/big.png"):p(120, 69):to(sprite, -150):anchor(cc.p(1, 0))
    A.line({
        {"union",{
            {"moveBy", 0.15, cc.p(-30, 0)},
            {"fadeTo", 0.15, 110},
        }},
        {"remove"},
    }):at(bigSnow1)
    A.line({
        {"union",{
            {"moveBy", 0.15, cc.p(-30, 0)},
            {"fadeTo", 0.15, 110},
        }},
        {"remove"},
    }):at(bigSnow2)
end

-- 移除雪痕迹
function M:snowStrailRemove(sprite, dis)
	if sprite._topSnow and not sprite._topSnow._willRemove then
        sprite._topSnow._willRemove = true
        sprite._topSnow:stop()
        sprite._bottomSnow:stop()
        if dis > 12 then
            -- 大片雪
            local bigSnow = D.img("biathlon/snowstrail/big.png"):p(140, 69):to(sprite, 150):anchor(cc.p(1, 0))
            A.line({
                {"union",{
                    {"moveBy", 0.15, cc.p(-40, 0)},
                    {"fadeTo", 0.15, 110},
                }},
                {"remove"},
            }):at(bigSnow)
        end
        A.line({
            {"fn", function()
                A.line({
                    {"imagerange", "biathlon/snowstrail/bottom/", 7, 2, 0.04},
                    {"remove"},
                }):at(sprite._bottomSnow) 
            end},
            {"imagerange", "biathlon/snowstrail/bottom/", 7, 2, 0.04},
            {"fn", function()
                sprite._topSnow = nil
            end},
            {"remove"},
        }):at(sprite._topSnow)
    end
end

-- 刀痕
function M:loadBlade(sprite, x, y, dis, velocity)
    if tolua.isnull(sprite) then return end
	-- 残影效果
    -- 参数: 1.间隐的时间, 2.间隐片断的大小, 3.贴图的宽高, 4.颜色值RGB, 5.贴图的路径或者贴图对象 
    if not sprite._blade1 then
        sprite._blade1 = CCMotionStreak:create(0.27, 5, 5, ccc3(255, 255, 255), "biathlon/trailing/2.png")
        sprite._blade2 = CCMotionStreak:create(0.35, 5, 5, ccc3(255, 255, 255), "biathlon/trailing/2.png")
        self:addChild(sprite._blade1, 999)
        self:addChild(sprite._blade2, 999)
    else
        if dis > 7 and dis <= 15 then return end
        if dis > 15  or velocity.x < 0 then
            sprite._blade1:reset()
		    sprite._blade2:reset()
        end
        sprite._blade1:setPosition(CCPointMake(x+15, y+7))
        sprite._blade2:setPosition(CCPointMake(x+15, y-3))
    end
end

-- 角色角度处理
function M:roleAngleDeal(spr, k)
    local sxV   = k > 0 and 1.3 or 4
    local angle = spr._ignoreControl == false and (-math.deg(math.atan(k))/sxV) or 0
    angle       = spr._ignoreSynPhyPos == false and angle or spr._defaultAngle
    -- A.one({"rotateTo", 1/60, angle}):at(spr)
    spr:rotate(angle)
end

-- 影子角度处理
function M:shadowAngleDeal(spr, body, dis, k)
    local shadow = spr:getBone(spr._name.."yy")

    dis = (dis-3)                             -- 跳跃高度过低时不做处理
    local opa = 255-dis*1.5                   -- 透明度
    local scY = 1-dis*0.002                   -- 缩放Y
    local scX = 1-(dis+math.abs(k)*150)*0.002 -- 缩放X
    opa = opa >= 0 and opa or 0
    scY = scY >= 0 and scY or 0
    scX = scX >= 0 and scX or 0
    if spr._hideShadow then
        opa = 0
    end
    -- 修改影子骨头纹理透明度
    spr:setBoneOpacity(spr._name.."yy", opa)
    local sxV = k > 0 and 5 or 3 -- 调整出的较为合适的两个值,使角色角度不完全垂直于赛道

    local angle = -math.deg(math.atan(k))/sxV 
    shadow:scaleY(scY):py(-dis):rotate(angle):scaleX(scX)
end

-- 精灵状态改变
function M:sprStateChange(spr, dis, body, velocity)
    if not spr._floorDis then spr._floorDis = dis end   

    if dis >= 7 and not spr._isCrash then
        if (dis - spr._floorDis) >= 2.5 then
            if spr._blade then
                spr._blade:reset()
            end
            spr._fsm:doEvent("toFly")-- 飞起
        end
    elseif dis < 7 then
        spr._fsm:doEvent("toStopFly")-- 停止飞起
        if spr._bellowAct then -- 如果有弹簧旋转动作, 关闭
            spr:stopAction(spr._bellowAct)
            spr._bellowAct = nil
            -- 弹簧弹飞效果结束
            spr:flyWithBellowsEffectEnd()
        end
    end
    spr._floorDis = dis -- 距离地面距离

    if spr._isPlayer then
        if math.abs(velocity.x) < 1 and math.abs(velocity.y) < 1 then
            spr._fsm:doEvent("toStop")-- 停止
        end
    end
end

-- 屏幕晃动判断
function M:screenShakeCheck(spr, dis)
    if dis > 15 and not spr._isCrash then
        spr._leaveFloorTime = spr._leaveFloorTime + 1 -- 离地时间
    elseif dis < 15 then
        if spr._isPlayer and spr._leaveFloorTime > 30 then
            -- 屏幕晃动
            self._mapManager:screenShake(8,  8)
        end
        spr._leaveFloorTime = 0
    end
end

-- 获取世界
function M:getWorld()
    return self._world
end

-- 移除物理世界
function M:removeWorld()
    if not self._world then return end
    -- 关闭实时跟随
    -- self:closeSC()
    -- 摧毁世界
    ZBox2D.destroyWorld(self._world)
    self._world = nil
end

----------------------
-- 结点析构
----------------------
function M:onDestructor()
	-- [超类调用]
	M.super.onDestructor(self)
end

----------------------
-- 过度动画开始之前的调用
----------------------
--[[--

1.资源释放，退出场景前的操作准备
2.处理视图结点卸载、事件解除绑定等相关操作

]]
function M:onExitTransitionStart()
    -- 移除物理世界
    self:removeWorld()
end


----------------------
-- 验证
----------------------
-- 验证合法性[构造函数参数]
function M:assertParameters(params)
    
end


return M