--[[
   Copyright (c) 2012-2021 baby-bus.com

   TODO:    赛道管理器
   Author:  Neighbour Mr.Wang
   Date:    2021.08.05
   
   http://www.baby-bus.com/
   
]]
local M     = classNode("TrackManager")


----------------------
-- 公共参数
----------------------
-- [常量]
-- 赛道偏移量Y
local trackOffsetY = {
    60, 145, 230, 315
}
-- 终点时偏移量Y(按名次)
local endLineOffsetY = {
    200, 50, 125, 280
}

local Config = require("app.biathlon.Config")

----------------------
-- 构造方法
----------------------
function M:ctor(params)
    -- [超类调用]
    M.super.ctor(self, params)

    -- 角色列表
    self._roleList      = params.roleList
    -- 赛道角色列表
    self._trackRoleList = {{}, {}, {}, {}}
    -- 地图管理器
    self._mapManager    = params.mapManager
    -- 到达终点的人数
    self._passGameNum   = 0
end

--------------------------------------------
-- 结点渲染
--------------------------------------------
--[[--

实体渲染，处理实体结点加载、事件绑定等相关操作

]]
function M:onRender()
    -- [超类调用]
    M.super.onRender(self)

    -- 安排赛道
    self:arrangeTrack()
end

-- 创建障碍物
function M:createBarrier()
    -- 障碍物生成点
    for k,pos in pairs(Config.barrierPoints) do
        local offYId  = N.random(1, 4)
        local offY    = trackOffsetY[offYId]
        local realPos = cc.p(pos.x, pos.y + offY) -- 实际的位置
        T.shuffle(Config.barrierIdList[k], 1)

        -- 障碍物
        local barrier = import("app.biathlon.node.Barrier").new({
            enterPos 	 = cc.p(500 + X_OFFSET, 50),
            types        = Config.barrierIdList[k][1],  -- 1冰柱2弹簧3雪坑4冰块
            offYId       = offYId,
            offY         = offY,
            mapManager   = self._mapManager,
        }):p(realPos)
        :to(self._mapManager)

        -- 获得离地面高度及地面斜率
        local dis,k = self._mapManager:getHighAndSlope(barrier, realPos.x, realPos.y)
        local angle = -math.deg(math.atan(k))
        barrier:rotate(angle)
    end 
end

-- 创建道具盒子
function M:createPropBox()
    -- 道具盒子生成点
    for k,pos in pairs(Config.propboxesPoints) do
        local offYId  = N.random(1, 4)
        local offY    = trackOffsetY[offYId]

        local realPos = cc.p(pos.x, pos.y + offY) -- 实际的位置
        T.shuffle(Config.propIdList[k], 1)

        -- 道具盒子
        local propbox = import("app.biathlon.node.PropBox").new({
            enterPos   = cc.p(500 + X_OFFSET, 50),
            offY       = offY,
            mapManager = self._mapManager,
            types      = Config.propIdList[k][1], --  1香蕉皮 2 红烧章鱼 3彩虹 4火箭
            allowTypes = Config.propIdList[k]     -- 允许出现的道具类型
        }):p(realPos.x, realPos.y-20)
        :to(self._mapManager)

        -- 获得离地面高度及地面斜率
        local dis,k = self._mapManager:getHighAndSlope(self, realPos.x, realPos.y)
        local angle = -math.deg(math.atan(k))
        propbox:rotate(angle)
    end 
end

-- 安排赛道
function M:arrangeTrack()
    local roleList = {}
    T.copyTable(roleList, self._roleList)
    T.shuffle(roleList)

    local bodyPos = cc.p(280, 580)-- 刚体位置

    for i,role in ipairs(roleList) do
        if #roleList == 1 then -- 只有一个, 说明在引导
            i               = 2 -- 安排到第二个赛道
            bodyPos         = cc.p(23500, 610)
            role._gameStage = 2 -- 防止直接触发第一次的打靶
        end
        
        table.insert(self._trackRoleList[i], role)
        -- 相对刚体的Y轴偏移量
        role._relBodyOffsetY = trackOffsetY[i]
        role._currentTrack   = i
        -- 创建body
        role:createBody(bodyPos)
        -- 记录初始赛道
        role._initTrack = i
    end
end

-- 更换赛道
function M:changeTrack(role, dir)
    if role._isChangeTrack then  return end
    if not role:canChangeTrack() then return end-- 能否切换赛道

    if (not role._ignoreControl) and (not role._ignoreSynPhyPos) then
	    local targetIndex   = role._currentTrack + dir
        if targetIndex > 4 or targetIndex <= 0 then return end

        T.removeOrder(self._trackRoleList[role._currentTrack], role)
        table.insert(self._trackRoleList[targetIndex], role)

        -- 目标y偏移量
        local targetOffY    = trackOffsetY[targetIndex] 

        -- 执行换赛道动作
        self:doChangeTrackAct(role, targetOffY, targetIndex)
    end
end

-- 强行设置角色y偏移量
function M:forceSetRoleOffsetY(role, offYId)
	-- 停止换赛道, 动作
    self:stopChangeTrackAct(role)

    T.removeOrder(self._trackRoleList[role._currentTrack], role)
    table.insert(self._trackRoleList[offYId], role)

    role._relBodyOffsetY = trackOffsetY[offYId]
    role._currentTrack   = offYId
end

-- 执行换赛道动作
function M:doChangeTrackAct(role, targetOffY, trackId, offsetY)
    offsetY = ifnil(offsetY, 8) -- 单次改变的值
    role._isChangeTrack = true
    -- 当前施加的y偏移量
    local curOffY       = role._relBodyOffsetY
    if targetOffY == curOffY then return end
    local dir           = (targetOffY - curOffY)/math.abs(targetOffY - curOffY)

    role._changeTrackAct = A.cycle({
        {"fn", function()
            role._relBodyOffsetY = role._relBodyOffsetY + dir*offsetY

            if math.abs(role._relBodyOffsetY - targetOffY) <= offsetY+1 then
                -- 停止换赛道, 动作
                self:stopChangeTrackAct(role)
                role._currentTrack = trackId
            end
        end},
        {"delay", 1/60},
    }):at(role)
end

-- 停止换赛道, 动作
function M:stopChangeTrackAct(role)
    if role._isChangeTrack then
        role:stopAction(role._changeTrackAct)
        role._isChangeTrack = false
    end
end

-- 换赛道冲线
function M:changeTrackToEndLine(role)
    if role._hasRoleCrashLine then return end
    role._hasRoleCrashLine = true

    self._passGameNum      = self._passGameNum + 1
    role._rank             = self._passGameNum -- 记录名次
    local v                = role._body:GetLinearVelocity()
    local vx               = 18 - self._passGameNum
    role._body:SetLinearVelocity(b2Vec2(vx, 0))

    -- 目标y偏移量
    local targetOffY    = endLineOffsetY[self._passGameNum]
    -- 停止换赛道, 动作
    self:stopChangeTrackAct(role)
    -- 执行换赛道动作
    self:doChangeTrackAct(role, targetOffY, 2)
end

-- 换赛道打靶
function M:changeTrackToShoot(role)
    -- 停止换赛道, 动作
    self:stopChangeTrackAct(role)

    -- 目标赛道id 为初始时的赛道
    local targetTrackId = role._initTrack
    role._shootTrack    = targetTrackId
    -- -- 目标y偏移量
    local targetOffY    = trackOffsetY[targetTrackId]

    T.removeOrder(self._trackRoleList[role._currentTrack], role)
    table.insert(self._trackRoleList[targetTrackId], role)

    -- 当前施加的y偏移量
    local curOffY       = trackOffsetY[role._currentTrack]
    local dir           = (targetOffY - curOffY)/math.abs(targetOffY - curOffY)

    -- 执行换赛道动作
    self:doChangeTrackAct(role, targetOffY, targetTrackId)
end

----------------------
-- 结点析构
----------------------
--[[--

实体析构，处理实体结点卸载、事件解除绑定等相关操作

]]
function M:onDestructor()
    -- [超类调用]
    M.super.onDestructor(self)
end








return M